import 'dart:async';

import 'package_exception.dart';

/**
 * Withing your custom package use `ErrorStream` mixin to get access to `error` stream 
 * and `recordError` function for method that must return value, if return type is 
 * void (or it doesn't matter), `handleError` or `handleErrorAsync` function can be 
 * used to auto handle an error or exception.
 *
 * `error` stream is of type `PackageException`. To use it within your app subscribe 
 * to `error` stream and get `Exception` or `Error` like that:
 * ```dart
 * pluginOrPackage.error.listen(
 *   (PackageException pkgExc) {
 *     var current = pkgExc.currentException; // Get current exception
 *     var exc = pkgExc.innerException; // Get an exception directly
 *     var error = pkgExc.innerError; // Get an error directly
 *     var stackTrace = pkgExc.stackTrace; // Get StackTrace if it have been provided
 *
 *     // do something with error or exception
 *   }
 * );
 * ```
 */
mixin ErrorStream {
  var _errorController = StreamController<PackageException>();

  /// Stream of `error`s
  Stream<PackageException> get error => _errorController.stream;

  /// Might be used for reinitializing `StreamController` after `disposeErrorStream` called
  initErrorStream() {
    if (_errorController == null) {
      _errorController = StreamController<PackageException>();
    }
  }

  /// Used to handle error in `synchronous` methods if return type is `void` or it doesn't matter
  void handleError(void func()) {
    try {
      func();
    } catch (e, stackTrace) {
      recordError(e, stackTrace);
    }
  }

  /// Used to handle error in `asynchronous` methods if return type is `void` or it doesn't matter
  void handleErrorAsync(Future future) =>
      future.catchError((e, stackTrace) => recordError(e, stackTrace));

  /// Wraps exception or error to `PackageException` and adding it to `error`s `StreamController`
  void recordError(e, [StackTrace stackTrace]) {
    if (e is Exception)
      _errorController.add(PackageException.fromException(e, stackTrace));
    else if (e is Error)
      _errorController.add(PackageException.fromError(e, stackTrace));
  }

  /// Dispose method for closing `StreamController`
  void disposeErrorStream() {
    _errorController.close();
  }
}
